#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ex:ts=4:sw=4:sts=4:et
# -*- tab-width: 4; c-basic-offset: 4; indent-tabs-mode: t -*-
#
# Copyright (C) 2010 Marco Pantaleoni. All rights reserved
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 2 as
#    published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__author__    = "Marco Pantaleoni"
__copyright__ = "Copyright (C) 2010 Marco Pantaleoni"
__license__   = "GPL v2"

import os
import sys
import argparse

from automa.tasks import task, Task, TaskContext
from automa.opts import options
import automa.virtual

PROGNAME = 'Automa'

def usage():
    sys.stderr.write("%s TASKNAME\n" % sys.argv[0])

def load_module(pathname):
    """Dynamically load a Python module."""
    import imp
    import traceback
    try:
        import hashlib
        m = hashlib.md5()
        m.update(pathname)
        mhash = m.hexdigest()
    except ImportError:
        import md5
        mhash = md5.new(pathname).hexdigest()
    f_in = open(pathname, 'rb')
    try:
        py_mod = imp.load_source(mhash, pathname, f_in)
    except ImportError, x:
        traceback.print_exc(file = sys.stderr)
        raise
    except:
        traceback.print_exc(file = sys.stderr)
        raise
    finally:
        try: fin.close()
        except: pass
    return py_mod

def main():
    """Program entry point."""
    import cmdutils

    def run_named_task(taskname, args):
        task = Task.TaskFromName(taskname)
        ctxt = TaskContext(task, cmdline_args=args)
        ctxt.run()

    log_level_choices = [
        ('VERBOSE_DEBUG', cmdutils.log.Logger.VERBOSE_DEBUG, "very verbose messages"),
        ('DEBUG', cmdutils.log.Logger.DEBUG, "debug messages"),
        ('INFO', cmdutils.log.Logger.INFO, "informational messages"),
        ('NOTIFY', cmdutils.log.Logger.NOTIFY, "more useful information messages"),
        ('WARN', cmdutils.log.Logger.WARN, "warning"),
        ('ERROR', cmdutils.log.Logger.ERROR, "an error"),
        ('FATAL', cmdutils.log.Logger.FATAL, "a fatal error")
    ]

    parser = argparse.ArgumentParser(prog=PROGNAME)
    parser.set_defaults(logger = None)
    parser.add_argument('--file', '-f', dest='file',
                        default='automate.py',
                        help='parse this file for task definitions')
    parser.add_argument('--list', action='store_true',
                        help='list tasks')
    parser.add_argument('--dry-run', '--dryrun', '--dry', '-n', dest='dry_run',
                        action='store_true',
                        help='simulate: don\'t execute commands and actions')
    parser.add_argument('--ignore-errors', '--ignoreerrors', dest='ignore_errors',
                        action='store_true',
                        help='don\'t bail out on errors (except fatal ones)')
    parser.add_argument('--loglevel', dest='loglevel',
                        choices=[l[0] for l in log_level_choices],
                        default='NOTIFY',
                        help='log level')
    subparsers = parser.add_subparsers(help='task help')

    load_module(os.path.join(os.path.abspath(os.getcwd()), 'automate.py'))

    for task in Task.Tasks():
        assert isinstance(task, Task)
        help_text = task.__doc__ or '%s task' % task.name
        subparser = subparsers.add_parser(task.name, help=help_text)
        subparser.set_defaults(func=lambda args, n=task.name: run_named_task(n, args))

    args = parser.parse_args()

    level = cmdutils.log.Logger.NOTIFY
    for ll in log_level_choices:
        if ll[0] == args.loglevel:
            level = ll[1]
            break
    logger = cmdutils.log.Logger([(level, sys.stderr)])
    args.logger = logger
    options.logger        = logger
    options.args          = args
    options.dry_run       = args.dry_run
    options.ignore_errors = args.ignore_errors

    # execute the task function
    args.func(args)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import os
import sys
import shutil
import fnmatch

try:
    import jinja2
    _has_jinja2 = True
except ImportError:
    _has_jinja2 = False

from automa import log

def _guess_autoescape(template_name):
    if template_name is None or '.' not in template_name:
        return False
    ext = template_name.rsplit('.', 1)[1]
    return ext in ('html', 'htm', 'xml')

if _has_jinja2:
    class _TemplatePathLoader(jinja2.BaseLoader):
        def __init__(self, encoding='utf-8'):
            self.encoding = encoding

        def get_source(self, environment, template):
            if not os.path.exists(template):
                raise jinja2.TemplateNotFound(template)
            mtime = os.path.getmtime(template)
            with file(template) as f:
                source = f.read().decode(self.encoding)
            def uptodate():
                try:
                    return os.path.getmtime(template) == mtime
                except OSError:
                    return False
            return source, template, uptodate

class Path(object):
    def __init__(self, pathname_or_path):
        if isinstance(pathname_or_path, Path):
            self.pathname = pathname_or_path.pathname
        else:
            self.pathname = pathname_or_path

    def __str__(self):
        return self.pathname

    def __repr__(self):
        return "<Path pathname:%r>" % self.pathname

    # ------------------------------------------------------------------------
    #   os.path wrappers
    # ------------------------------------------------------------------------

    @property
    def abspath(self):
        return Path(os.path.abspath(self.pathname))

    @property
    def basename(self):
        return os.path.basename(self.pathname)

    @property
    def dirname(self):
        return Path(os.path.dirname(self.pathname))

    @property
    def exists(self):
        return os.path.exists(self.pathname)

    @property
    def lexists(self):
        return os.path.lexists(self.pathname)

    @property
    def expanduser(self):
        return Path(os.path.expanduser(self.pathname))

    @property
    def expandvars(self):
        return Path(os.path.expandvars(self.pathname))

    @property
    def atime(self):
        return os.path.getatime(self.pathname)
    @property
    def ctime(self):
        return os.path.getctime(self.pathname)
    @property
    def mtime(self):
        return os.path.getmtime(self.pathname)

    @property
    def size(self):
        return os.path.getsize(self.pathname)

    @property
    def isabs(self):
        return os.path.isabs(self.pathname)

    @property
    def isfile(self):
        return os.path.isfile(self.pathname)

    @property
    def isdir(self):
        return os.path.isdir(self.pathname)

    @property
    def islink(self):
        return os.path.islink(self.pathname)

    @property
    def ismount(self):
        return os.path.ismount(self.pathname)

    @property
    def normcase(self):
        return Path(os.path.normcase(self.pathname))

    @property
    def normpath(self):
        return Path(os.path.normpath(self.pathname))

    @property
    def normalized(self):
        return Path(os.path.normcase(os.path.normpath(self.pathname)))

    def normalize(self):
        self.pathname = os.path.normcase(os.path.normpath(self.pathname))
        return self

    @property
    def realpath(self):
        return Path(os.path.realpath(self.pathname))

    @property
    def relpath(self):
        return Path(os.path.relpath(self.pathname))

    def split(self):
        t = os.path.split(self.pathname)
        return (Path(t[0]), t[1])

    def join(self, other):
        if isinstance(other, Path):
            return Path(os.path.join(self.pathname, other.pathname))
        return Path(os.path.join(self.pathname, other))

    def splitext(self):
        t = os.path.splitext(self.pathname)
        return (Path(t[0]), t[1])

    @property
    def ext(self):
        t = os.path.splitext(self.pathname)
        ext = t[1]
        if ext and ext.startswith('.'):
            ext = ext[1:]
        return ext

    def samefile(self, other):
        if not isinstance(other, Path):
            other = Path(other)
        return os.path.samefile(self.pathname, other.pathname)

    def same(self, other):
        if not isinstance(other, Path):
            other = Path(other).normalize()
        this = self.normalized
        return (this.pathname == other.pathname)

    def __eq__(self, other):
        if isinstance(other, Path):
            return self.pathname == other.pathname
        return False

    def __add__(self, other):
        return self.join(other)

    def __div__(self, other):
        return self.join(other)

    def __abs__(self):
        return self.abspath

    def __coerce__(self, other):
        if isinstance(other, basestring):
            return (self, Path(other))
        elif isinstance(other, Path):
            return (self, other)
        return None

    def __iter__(self):
        fh = self.open()
        return fh

    # ------------------------------------------------------------------------
    #   os wrappers
    # ------------------------------------------------------------------------

    def chdir(self):
        try:
            os.chdir(self.pathname)
        except Exception, exc:
            log.error("executing os.chdir(%r)" % self.pathname)
        return self

    def access(self, mode):
        return os.access(self.pathname, mode)

    def chroot(self):
        try:
            os.chroot(self.pathname)
        except Exception, exc:
            log.error("executing os.chroot(%r)" % self.pathname)
        return self

    def chflags(self, flags):
        try:
            os.chflags(self.pathname, flags)
        except Exception, exc:
            log.error("executing os.chflags(%r, %r)" % (self.pathname, flags))
        return self

    def lchflags(self, flags):
        try:
            os.lchflags(self.pathname, flags)
        except Exception, exc:
            log.error("executing os.lchflags(%r, %r)" % (self.pathname, flags))
        return self

    def chmod(self, mode):
        try:
            os.chmod(self.pathname, mode)
        except Exception, exc:
            log.error("executing os.chmod(%r, %r)" % (self.pathname, mode))
        return self

    def lchmod(self, mode):
        try:
            os.lchmod(self.pathname, mode)
        except Exception, exc:
            log.error("executing os.lchmod(%r, %r)" % (self.pathname, mode))
        return self

    def chown(self, uid, gid):
        try:
            os.chown(self.pathname, uid, gid)
        except Exception, exc:
            log.error("executing os.chown(%r, %r, %r)" % (self.pathname, uid, gid))
        return self

    def lchown(self, uid, gid):
        try:
            os.lchown(self.pathname, uid, gid)
        except Exception, exc:
            log.error("executing os.lchown(%r, %r, %r)" % (self.pathname, uid, gid))
        return self

    def link(self, link_name):
        try:
            os.link(self.pathname, link_name)
        except Exception, exc:
            log.error("executing os.link(%r, %r)" % (self.pathname, link_name))
        return self

    def listdir(self):
        try:
            l = os.listdir(self.pathname)
        except Exception, exc:
            log.error("executing os.listdir(%r)" % self.pathname)
        else:
            for entry in l:
                yield entry

    def mkdir(self, mode=0777):
        try:
            os.mkdir(self.pathname, mode)
        except Exception, exc:
            log.error("executing os.mkdir(%r, %r)" % (self.pathname, mode))
        return self

    def makedirs(self, mode=0777):
        try:
            os.makedirs(self.pathname, mode)
        except Exception, exc:
            log.error("executing os.makedirs(%r, %r)" % (self.pathname, mode))
        return self

    @property
    def readlink(self):
        try:
            return os.readlink(self.pathname)
        except Exception, exc:
            log.error("executing os.readlink(%r)" % self.pathname)
        return None

    def remove(self):
        try:
            os.remove(self.pathname)
        except Exception, exc:
            log.error("executing os.remove(%r)" % self.pathname)
        return self

    def rename(self, dst):
        if not isinstance(dst, Path):
            dst = Path(dst)
        try:
            os.rename(self.pathname, dst.pathname)
        except Exception, exc:
            log.error("executing os.rename(%r, %r)" % (self.pathname, dst.pathname))
        return self

    def renames(self, dst):
        if not isinstance(dst, Path):
            dst = Path(dst)
        try:
            os.renames(self.pathname, dst.pathname)
        except Exception, exc:
            log.error("executing os.renames(%r, %r)" % (self.pathname, dst.pathname))
        return self

    def rmdir(self):
        try:
            os.rmdir(self.pathname)
        except Exception, exc:
            log.error("executing os.rmdir(%r)" % self.pathname)
        return self

    @property
    def stat(self):
        try:
            return os.stat(self.pathname)
        except Exception, exc:
            log.error("executing os.stat(%r)" % self.pathname)
        return None

    @property
    def lstat(self):
        try:
            return os.lstat(self.pathname)
        except Exception, exc:
            log.error("executing os.stat(%r)" % self.pathname)
        return None

    def symlink(self, link_name):
        try:
            os.symlink(self.pathname, link_name)
        except Exception, exc:
            log.error("executing os.symlink(%r, %r)" % (self.pathname, link_name))
        return self

    def unlink(self):
        try:
            os.unlink(self.pathname)
        except Exception, exc:
            log.error("executing os.unlink(%r)" % self.pathname)
        return self

    # ------------------------------------------------------------------------
    #   file read & write helpers
    # ------------------------------------------------------------------------

    def read(self, mode='rb', encoding=None):
        try:
            fh = open(self.pathname, mode)
        except Exception, exc:
            log.error("executing open(%r, %r)" % (self.pathname, mode))
            return ""
        r = fh.read()
        fh.close()
        if encoding is not None:
            r = r.decode(encoding)
        return r

    def readlines(self, mode='r', encoding='utf-8'):
        try:
            fh = open(self.pathname, mode)
        except Exception, exc:
            log.error("executing open(%r, %r)" % (self.pathname, mode))
            return []
        r = fh.readlines()
        fh.close()
        if encoding is not None:
            r = [line.decode(encoding) for line in r]
        return r

    def write(self, text, mode='wb'):
        try:
            fh = open(self.pathname, mode)
        except Exception, exc:
            log.error("executing open(%r, %r)" % (self.pathname, mode))
            return self
        fh.write(text)
        fh.close()
        return self

    def writelines(self, lines, mode='w'):
        return self.write("\n".join(lines), mode=mode)

    def open(self, mode='r'):
        try:
            return open(self.pathname, mode)
        except Exception, exc:
            log.error("executing open(%r, %r)" % (self.pathname, mode))
        return None

    # ------------------------------------------------------------------------
    #   shutil wrappers and various helpers
    # ------------------------------------------------------------------------

    def copy(self, dst):
        if not isinstance(dst, Path):
            dst = Path(dst)
        try:
            shutil.copy(self.pathname, dst.pathname)
        except Exception, exc:
            log.error("executing shutil.copy(%r, %r)" % (self.pathname, dst.pathname))
        return self

    def copy_with_perms(self, dst):
        if not isinstance(dst, Path):
            dst = Path(dst)
        try:
            shutil.copy2(self.pathname, dst.pathname)
        except Exception, exc:
            log.error("executing shutil.copy2(%r, %r)" % (self.pathname, dst.pathname))
        return self

    def copy_contents(self, dst):
        if not isinstance(dst, Path):
            dst = Path(dst)
        try:
            shutil.copyfile(self.pathname, dst.pathname)
        except Exception, exc:
            log.error("executing shutil.copyfile(%r, %r)" % (self.pathname, dst.pathname))
        return self

    def move(self, dst):
        if not isinstance(dst, Path):
            dst = Path(dst)
        try:
            shutil.move(self.pathname, dst.pathname)
        except Exception, exc:
            log.error("executing shutil.move(%r, %r)" % (self.pathname, dst.pathname))
        return self

    def copytree(self, dst, symlinks=True, ignore=None):
        if not isinstance(dst, Path):
            dst = Path(dst)
        try:
            shutil.copytree(self.pathname, dst.pathname, symlinks=symlinks,
                            ignore=ignore)
        except Exception, exc:
            log.error("executing shutil.copytree(%r, %r)" % (self.pathname, dst.pathname))
        return self

    def rmtree(self, ignore_errors=False):
        try:
            shutil.rmtree(self.pathname)
        except Exception, exc:
            log.error("executing shutil.rmtree(%r)" % self.pathname)
        return self

    # ------------------------------------------------------------------------
    #   various helpers
    # ------------------------------------------------------------------------

    def startswith(self, text):
        return self.pathname.startswith(text)
    def endswith(self, text):
        return self.pathname.endswith(text)

    def walkfiles(self, pattern="*"):
        try:
            l = os.listdir(self.pathname)
        except Exception, exc:
            log.error("executing os.listdir(%r)" % self.pathname)
        else:
            for entry in l:
                if fnmatch.fnmatch(entry, pattern):
                    yield Path(os.path.join(self.pathname, entry))

    @classmethod
    def _get_template(cls, template_path, encoding='utf-8'):
        from opts import options
        if not _has_jinja2:
            log.error("jinja2 is not installed")
            return None
        if not isinstance(template_path, Path):
            template_path = Path(template_path)
        loader = _TemplatePathLoader(encoding=encoding)
        environment = jinja2.Environment(autoescape=_guess_autoescape,
                                         loader=loader,
                                         extensions=['jinja2.ext.autoescape'])
        environment.globals = options
        return environment.get_template(template.pathname)

    def render_to_string(self, context={}, encoding='utf-8'):
        tpl = Path._get_template(self, encoding)
        return tpl.render(context)

    def render_from_string(self, template, context={}, encoding='utf-8'):
        from opts import options
        if not _has_jinja2:
            log.error("jinja2 is not installed")
            return None
        tpl = jinja2.Template(template,
                              autoescape=_guess_autoescape,
                              extensions=['jinja2.ext.autoescape'])
        tpl.globals = options
        return self.write(tpl.render(context))

    def render_to(self, dst, context={}, encoding='utf-8'):
        if not isinstance(dst, Path):
            dst = Path(dst)
        tpl = Path._get_template(self, encoding)
        return dst.write(tpl.render(context))

    def render_from(self, template, context={}, encoding='utf-8'):
        tpl = Path._get_template(template, encoding)
        return self.write(tpl.render(context))



# -*- coding: utf-8 -*-

options = None

class Options(dict):
    """
    An extended dictionary, allowing access to keys also as instance attributes.

       >>> d = Options()

       >>> d['k1'] = 23

       >>> d['k1']
       23

       >>> d.k1
       23

       >>> d.var = 'hello'

       >>> d['var']
       'hello'

       >>> d.var
       'hello'

       >>> d = Options(initval = 5)
       >>> d.initval
       5
    """

    def __getattr__(self, key):
        return self.get(key, None)

    def __setattr__(self, key, value):
        self[key] = value

    def _getFQ(self, key):
        d = self
        while '.' in key:
            (prefix, postfix) = key.split('.', 1)
            v = d.get(prefix, None)
            if v is None:
                return None
            d = v
            key = postfix
        return d.get(key, None)

    def __call__(self, **kwargs):
        self.update(kwargs)
        return self

def _setup_default_options():
    global options

    import sys
    import platform

    if options is not None:
        return options

    vtuple = platform.python_version_tuple()
    options = Options(sys = Options(executable          = sys.executable,
                                    prefix              = sys.prefix,
                                    platform            = sys.platform,
                                    machine             = platform.machine(),
                                    system              = platform.system(),
                                    python_version      = vtuple,
                                    python_version_tag  = '%s.%s' % (vtuple[0], vtuple[1]),
                                    python_version_info = sys.version_info,
                                    lib_build_ver       = '%s-%s' % (platform.system().lower(), platform.machine())),
                      dry_run = False
                      )
    return options

_setup_default_options()

def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ex:ts=4:sw=4:sts=4:et
# -*- tab-width: 4; c-basic-offset: 4; indent-tabs-mode: t -*-
#
# Copyright (C) 2010 Marco Pantaleoni. All rights reserved
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 2 as
#    published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__author__    = "Marco Pantaleoni"
__copyright__ = "Copyright (C) 2010 Marco Pantaleoni"
__license__   = "GPL v2"

import sys
import traceback
import datetime

def get_active_context():
    return TaskContext.Active()

def call_task(*args, **kwargs):
    cntxt = TaskContext.Active()
    cntxt.call_task(*args, **kwargs)

class Task(object):
    __doc__ = ""
    _tasks            = []
    _task_by_name     = {}
    _task_by_fullname = {}

    def __init__(self, func):
        self.func     = func
        self.__name__ = func.__name__
        self.name     = func.__name__
        self.fullname = "%s.%s" % (func.__module__, func.__name__)
        if hasattr(func, '__doc__'):
            self.__doc__ = func.__doc__
        self.depends  = []

        self._started    = None
        self._completed  = None
        self._started_by = None

        self._tasks.append(self)
        self._task_by_name[self.name] = self
        self._task_by_fullname[self.fullname] = self

    def __call__(self, *args, **kwargs):
        ctxt = TaskContext(self, args, kwargs)
        return ctxt.run()

    def __str__(self):
        return '<Task %r>' % self.fullname

    def __repr__(self):
        return '<Task %r>' % self.fullname

    @classmethod
    def Tasks(cls):
        return list(cls._tasks)

    @classmethod
    def TaskFromShortName(cls, name):
        return cls._task_by_name.get(name, None)
    @classmethod
    def TaskFromFullName(cls, name):
        return cls._task_by_fullname.get(name, None)
    @classmethod
    def TaskFromName(cls, name):
        return cls.TaskFromFullName(name) or cls.TaskFromShortName(name)

def task(f):
    """@task decorator"""
    if isinstance(f, Task):
        return f
    return Task(f)

class depends(object):
    """@depends decorator"""
    def __init__(self, *args):
        self.args = args

    def __call__(self, f):
        if not isinstance(f, Task):
            f = Task(f)
        f.depends = self.args
        return f

class TaskContext(object):
    _active_task_context = None

    def __init__(self, task, args=None, kwargs=None, cmdline_args=None,
                 called_from=None, level=1):
        self.task   = task
        self.args   = args or []
        self.kwargs = kwargs or {}
        self.cmdline_args = cmdline_args

        self.called_from = called_from
        self.level       = level
        self._logger     = None

        self.executed    = []  # tasks which have been started
        self.completed   = []  # tasks which have been executed up to completion

    def __repr__(self):
        return '<TaskContext 0x%X task:%r>' % (id(self), self.task.fullname)

    def run(self):
        TaskContext._active_task_context = self
        self.notify("TASK %s CALLED (ctxt:0x%X)" % (self.task.name, id(self)), color='bold magenta', level=self.level-1)
        assert isinstance(self.task, Task)
        if len(self.task.depends) > 0:
            self.notify("DEPENDS ON %s" % str(self.task.depends), color='bold magenta')
            for item in self.task.depends:
                if isinstance(item, basestring):
                    item = Task.TaskFromName(item)
                if isinstance(item, Task):
                    if not self.task_is_needed(item):
                        self.notify("TASK %s IS NOT NEEDED" % item, color='bold magenta')
                        continue
                    self.notify("NEEDED %s BY %s" % (item, self.task), color='bold magenta')
                    ctxt = TaskContext(item, cmdline_args=self.cmdline_args)
                    ctxt._logger = self._logger
                    ctxt.called_from = self
                    ctxt.level = self.level + 1
                    ctxt.run()
        self.notify("TASK %s STARTED" % self.task.name, color='bold magenta', level=self.level-1)
        self.executed.append(self.task)
        if self.called_from:
            self.task._started_by = self.called_from.task
        self.task._started = datetime.datetime.now()
        rv = self.task.func(self)
        self.task._completed = datetime.datetime.now()
        self.completed.append(self.task)
        self.notify("TASK %s COMPLETED" % self.task.name, color='bold magenta', level=self.level-1)
        return rv

    def task_is_needed(self, task):
        assert isinstance(task, Task)
        return task._started is None

    def __call__(self, *args, **kwargs):
        self.args   = args or self.args
        self.kwargs = kwargs or self.kwargs
        return self.run()

    def call_task(self, taskname, *args, **kwargs):
        task = Task.TaskFromName(taskname)
        ctxt = TaskContext(task, args=args, kwargs=kwargs, cmdline_args=self.cmdline_args)
        ctxt._logger = self._logger
        ctxt.called_from = self
        ctxt.level = self.level + 1
        return ctxt.run()

    @classmethod
    def Active(cls):
        return TaskContext._active_task_context

    @property
    def logger(self):
        if self._logger:
            return self._logger
        if self.cmdline_args:
            self._logger = self.cmdline_args.logger
        else:
            import cmdutils, sys
            self._logger = cmdutils.log.Logger([(cmdutils.log.Logger.NOTIFY, sys.stderr)])
        return self._logger

    def _indent_text(self, text, amount, space="  "):
        indent = space * amount
        lines = text.split('\n')
        return '\n'.join([(indent + line) for line in lines])

    def debug(self, msg, color='cyan', level=None, extra_indent=0, *args, **kwargs):
        if level is None:
            level = self.level
        level += extra_indent
        return self.logger.debug(self._indent_text('DEBUG: ' + msg, level),  color=color)
    def info(self, msg, color=None, level=None, extra_indent=0, *args, **kwargs):
        if level is None:
            level = self.level
        level += extra_indent
        return self.logger.info(self._indent_text(msg, level), color=color)
    def notify(self, msg, color='green', level=None, extra_indent=0, *args, **kwargs):
        if level is None:
            level = self.level
        level += extra_indent
        return self.logger.notify(self._indent_text(msg, level), color=color)
    def warn(self, msg, color='bold yellow', level=None, extra_indent=0, *args, **kwargs):
        if level is None:
            level = self.level
        level += extra_indent
        return self.logger.warn(self._indent_text('WARN: ' + msg, level), color=color)
    def error(self, msg, color='bold red', level=None, extra_indent=0, ignore_errors=None, *args, **kwargs):
        if level is None:
            level = self.level
        level += extra_indent
        if ignore_errors is None:
            if self.cmdline_args:
                ignore_errors = self.cmdline_args.ignore_errors
        prefix = 'ERROR: '
        if ignore_errors:
            prefix = 'ERROR (ignored): '
        text = prefix + msg
        exctype, value = sys.exc_info()[:2]
        if exctype is not None:
            text += "\nEXCEPTION\n%s" % traceback.format_exc()
        rv = self.logger.error(self._indent_text(text, level), color=color)
        if not ignore_errors:
            sys.exit(1)
        return rv
    def fatal(self, msg, color='bold red_bg', level=None, extra_indent=0, *args, **kwargs):
        if level is None:
            level = self.level
        level += extra_indent
        text = '\n\nFATAL: ' + msg
        exctype, value = sys.exc_info()[:2]
        if exctype is not None:
            text += "\nEXCEPTION\n%s" % traceback.format_exc()
        text += '\n\n'
        rv = self.logger.error(self._indent_text(text, level), color=color)
        sys.exit(1)
        return rv
